/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2016 OX Software GmbH.
 *
 * @author Ingo Schmidt-Rosbiegal <ingo.schmidt-rosbiegal@open-xchange.com>
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/model/mergecollection', [
    'io.ox/office/tk/utils',
    'io.ox/office/baseframework/model/modelobject',
    'io.ox/office/spreadsheet/utils/sheetutils'
], function (Utils, ModelObject, SheetUtils) {

    'use strict';

    // class MergeCollection ==================================================

    /**
     * Collects information about all merged cell ranges of a single sheet in a
     * spreadsheet document.
     *
     * Triggers the following events:
     * - 'insert:merged'
     *      After new merged ranges have been inserted into the collection.
     *      Event handlers receive the following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {Array} ranges
     *          An array with the addresses of all new merged ranges.
     *      (3) {Object} [options]
     *          Optional parameters:
     *          - {Boolean} [options.implicit=false]
     *              If set to true, the merged ranges have been inserted
     *              implicitly, for example while rows, columns, or cell ranges
     *              have been inserted or deleted in the sheet.
     * - 'delete:merged'
     *      After merged ranges have been deleted from the collection. Event
     *      handlers receive the following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {Array} ranges
     *          An array with the addresses of all deleted merged ranges.
     *      (3) {Object} [options]
     *          Optional parameters:
     *          - {Boolean} [options.implicit=false]
     *              If set to true, the merged ranges have been deleted
     *              implicitly, for example while rows, columns, or cell ranges
     *              have been inserted or deleted in the sheet.
     *
     * @constructor
     *
     * @extends ModelObject
     *
     * @param {SpreadsheetApplication} app
     *  The application that contains this collection instance.
     *
     * @param {SheetModel} sheetModel
     *  The sheet model instance containing this collection.
     */
    function MergeCollection(app, sheetModel) {

        var // self reference
            self = this,

            // the spreadsheet document model
            docModel = app.getModel(),

            // all merged ranges
            mergedRanges = [];

        // base constructor ---------------------------------------------------

        ModelObject.call(this, app);

        // private methods ----------------------------------------------------

        /**
         * Returns whether the single columns or rows of the passed range are
         * merged. The merged columns/rows must exactly fit into the range, and
         * all columns/rows in the range must be merged.
         *
         * @param {Object} range
         *  The address of a range.
         *
         * @returns {Boolean}
         *  Whether the columns/rows of the specified range are merged.
         */
        function isColRowMerged(range, columns) {

            var // all merged ranges that are contained in the passed range
                containedRanges = self.getMergedRanges(range, 'contain'),
                // the array index in cell addresses
                addrIndex = columns ? 0 : 1;

            // the number of merged ranges contained by the passed range must be equal to the
            // column/row count of the range, otherwise there is no need to iterate at all
            if (containedRanges.length !== SheetUtils.getIndexCount(range, columns)) {
                return false;
            }

            // check size and position of the merged ranges (if all merged ranges start and
            // end at the boundaries of the passed range, and are single-sized in the other
            // direction, the entire range is filled by single-sized merged ranges)
            return _.all(containedRanges, function (mergedRange) {
                return (mergedRange.start[addrIndex] === mergedRange.end[addrIndex]) &&
                    (mergedRange.start[1 - addrIndex] === range.start[1 - addrIndex]) &&
                    (mergedRange.end[1 - addrIndex] === range.end[1 - addrIndex]);
            });
        }

        /**
         * Triggers the specified event, if the passed array of range addresses
         * is not empty.
         */
        function triggerEvent(type, ranges, options) {
            if (ranges.length > 0) {
                self.trigger(type, ranges, options);
            }
        }

        /**
         * Recalculates the position of all merged ranges, after columns or
         * rows have been inserted into or deleted from the sheet.
         */
        function transformationHandler(interval, insert, columns) {

            var // all deleted merged ranges
                deletedRanges = [],
                // all inserted merged ranges
                insertedRanges = [];

            // visit all merged ranges in reversed order, to be able to delete in-place
            Utils.iterateArray(mergedRanges, function (mergedRange, index) {

                var // get the transformed range (will return null for resulting 1x1 ranges)
                    newRange = self.transformRange(mergedRange, interval, insert, columns);

                // handle deleted and changed merged ranges
                if (!newRange) {
                    deletedRanges.push(mergedRange);
                    mergedRanges.splice(index, 1);
                } else if (!_.isEqual(mergedRange, newRange)) {
                    insertedRanges.push(_.copy(newRange, true));
                    deletedRanges.push(mergedRange);
                    mergedRanges[index] = newRange;
                }
            }, { reverse: true });

            // notify listeners (with option 'implicit' due do implicit update while inserting columns/rows)
            triggerEvent('delete:merged', deletedRanges, { implicit: true });
            triggerEvent('insert:merged', insertedRanges, { implicit: true });
        }

        // protected methods --------------------------------------------------

        /**
         * Creates and returns a cloned instance of this collection for the
         * specified sheet.
         *
         * @internal
         *  Used by the class SheetModel during clone construction. DO NOT CALL
         *  from external code!
         *
         * @param {SheetModel} targetModel
         *  The model instance of the new cloned sheet that will own the clone
         *  returned by this method.
         *
         * @returns {MergeCollection}
         *  A clone of this collection, initialized for ownership by the passed
         *  sheet model.
         */
        this.clone = function (targetModel) {
            // construct a new collection, pass all own ranges as hidden parameter
            return new MergeCollection(app, targetModel, mergedRanges);
        };

        // public methods -----------------------------------------------------

        /**
         * Returns, whether the cell at the specified address is the upper left
         * corner of a merged cell range.
         *
         * @param {Number[]} address
         *  The address of a cell.
         *
         * @returns {Boolean}
         *  Whether the cell at the specified address is the upper left corner
         *  of a merged cell range.
         */
        this.isReferenceCell = function (address) {
            return this.getMergedRange(address, 'reference') !== null;
        };

        /**
         * Returns, whether the cell at the specified address is a hidden cell
         * inside a merged cell range.
         *
         * @param {Number[]} address
         *  The address of a cell.
         *
         * @returns {Boolean}
         *  Whether the cell at the specified address is a hidden cell inside a
         *  merged cell range.
         */
        this.isHiddenCell = function (address) {
            return this.getMergedRange(address, 'hidden') !== null;
        };

        /**
         * Returns, whether the cell at the specified address is part of a
         * merged cell. It can be the reference cell or a hidden cell.
         *
         * @param {Number[]} address
         *  The address of a cell.
         *
         * @returns {Boolean}
         *  Whether the cell is part of a merged cell range.
         */
        this.isMergedCell = function (address) {
            return this.getMergedRange(address, 'all') !== null;
        };

        /**
         * Returns whether the passed range is contained in this collection.
         *
         * @param {Object} range
         *  The address of a range.
         *
         * @returns {Boolean}
         *  Whether the specified range is contained in this collection.
         */
        this.isMergedRange = function (range) {
            return _.any(mergedRanges, function (mergedRange) { return _.isEqual(range, mergedRange); });
        };

        /**
         * Returns whether the single columns of the passed range are merged.
         * The merged columns must exactly fit into the range, and all columns
         * in the range must be merged.
         *
         * @param {Object} range
         *  The address of a range.
         *
         * @returns {Boolean}
         *  Whether the specified range is entirely vertically merged.
         */
        this.isVerticallyMerged = function (range) {
            return isColRowMerged(range, true);
        };

        /**
         * Returns whether the single rows of the passed range are merged. The
         * merged rows must exactly fit into the range, and all rows in the
         * range must be merged.
         *
         * @param {Object} range
         *  The address of a range.
         *
         * @returns {Boolean}
         *  Whether the specified range is entirely horizontally merged.
         */
        this.isHorizontallyMerged = function (range) {
            return isColRowMerged(range, false);
        };

        /**
         * Returns the range address of a merged cell range for a specified
         * cell address.
         *
         * @param {Number[]} address
         *  The address of a cell.
         *
         * @param {String} [type='all']
         *  The address type of the specified address. It can be an arbitrary
         *  cell inside a merged cell range (type === 'all'), or it can be
         *  restricted to the upper left reference cell (type === 'reference')
         *  or to a hidden cell (type === 'hidden'). A hidden cell is any cell
         *  inside the merged cell range, that is not the upper left reference
         *  cell. If this address type is not specified, or if it is not one of
         *  the three types 'all', 'reference' or 'hidden', it defaults to
         *  'all'.
         *
         * @returns {Object|Null}
         *  The range address of a merged range that covers the passed cell
         *  address; or null, if no merged range exists at the address.
         */
        this.getMergedRange = function (address, type) {

            var // the merged range containing the cell with the passed address
                mergedRange = SheetUtils.findFirstRange(mergedRanges, address);

            if (mergedRange) {
                switch (type) {
                case 'reference':
                    if (!_.isEqual(mergedRange.start, address)) { mergedRange = null; }
                    break;
                case 'hidden':
                    if (_.isEqual(mergedRange.start, address)) { mergedRange = null; }
                    break;
                }
            }

            return mergedRange ? _.copy(mergedRange, true) : null;
        };

        /**
         * Returns the range addresses of all merged cell ranges that overlap
         * with one of the passed ranges.
         *
         * @param {Object|Array} ranges
         *  A single cell range address, or an array with range addresses.
         *
         * @param {String} [type='all']
         *  If set to 'all' or omitted, returns all merged ranges that overlap
         *  the passed ranges with any cell. If set to 'contain', returns only
         *  merged ranges that are completely contained in any of the passed
         *  ranges. If set to 'reference', returns only merged ranges whose
         *  reference cell is contained in any of the passed ranges.
         *
         * @returns {Array}
         *  The range addresses of all merged ranges that overlap with one of
         *  the passed ranges.
         */
        this.getMergedRanges = function (ranges, type) {

            var // the merged ranges not matching any passed range yet
                pendingRanges = _.clone(mergedRanges),
                // the resulting merged ranges
                resultRanges = [];

            // process all passed ranges
            _.any(_.getArray(ranges), function (range) {

                // iterate backwards through pendingRanges to be able to remove array elements
                Utils.iterateArray(pendingRanges, function (mergedRange, index) {

                    // move matching merged range from pendingRanges to resultRanges
                    if (SheetUtils.rangeOverlapsRange(range, mergedRange)) {
                        switch (type) {
                        case 'reference':
                            if (SheetUtils.rangeContainsCell(range, mergedRange.start)) {
                                resultRanges.push(mergedRange);
                            }
                            break;
                        case 'contain':
                            if (SheetUtils.rangeContainsRange(range, mergedRange)) {
                                resultRanges.push(mergedRange);
                            }
                            break;
                        default:
                            resultRanges.push(mergedRange);
                        }
                        // remove ALL overlapping ranges from pendingRanges (regardless of passed type)
                        pendingRanges.splice(index, 1);
                    }
                }, { reverse: true });

                // early exit of the _.any() loop, if all merged ranges will be returned
                return pendingRanges.length === 0;
            });

            return resultRanges;
        };

        /**
         * Returns the merged range that completely contains or is equal to the
         * passed cell range. Note the different to the method
         * getMergedRanges() which returns all merged ranges that are contained
         * IN the passed ranges.
         *
         * @param {Object} range
         *  A range address.
         *
         * @returns {Object|Null}
         *  The address of a merged range that contains the passed range if
         *  existing, otherwise null
         */
        this.getBoundingMergedRange = function (range) {
            var boundingRange = _.find(mergedRanges, function (mergedRange) { return SheetUtils.rangeContainsRange(mergedRange, range); });
            return boundingRange ? boundingRange : null;
        };

        /**
         * Returns whether any of the passed cell ranges overlaps with at least
         * one merged cell range.
         *
         * @param {Object|Array} ranges
         *  Address of a single cell range, or an array of range addresses.
         *
         * @returns {Boolean}
         *  Whether any of the passed cell ranges contains or overlaps with at
         *  least one merged cell range.
         */
        this.rangesOverlapAnyMergedRange = function (ranges) {
            return SheetUtils.rangesOverlapRanges(ranges, mergedRanges);
        };

        /**
         * Expands the passed range address, so that it will include all merged
         * ranges it is currently covering partly.
         *
         * @param {Object} range
         *  The address of a cell range that will be expanded to the merged
         *  ranges contained in this collection.
         *
         * @returns {Object}
         *  The address of the expanded range.
         */
        this.expandRangeToMergedRanges = function (range) {

            var // flat copy of the merged ranges, to prevent processing them multiple times
                pendingRanges = _.clone(mergedRanges),
                // the length of the pending ranges array
                pendingLength = pendingRanges.length;

            // loop iterator function must be defined outside the do-while loop
            function expandToMergedRange(mergedRange, index) {
                // process all merged ranges that overlap the current range
                if (SheetUtils.rangeOverlapsRange(range, mergedRange)) {
                    range = SheetUtils.getBoundingRange(range, mergedRange);
                    pendingRanges.splice(index, 1);
                }
            }

            // loop as long as the range will still be expanded with other merged ranges
            if (pendingRanges.length > 0) {
                do {
                    // store the current length of the pending ranges array (checking length afterwards)
                    pendingLength = pendingRanges.length;
                    // iterate backwards to be able to remove an element from the array
                    Utils.iterateArray(pendingRanges, expandToMergedRange, { reverse: true });
                } while (pendingRanges.length < pendingLength);
            }

            // return the fully expanded range
            return _.copy(range, true);
        };

        /**
         * Shrinks the passed range address, so that it will not include any
         * merged ranges it is currently covering partly.
         *
         * @param {Object} range
         *  The address of a cell range that will be shrunk until it does not
         *  partly contain any merged ranges.
         *
         * @param {String} direction
         *  The direction in which the passed range will be shrunk. Supported
         *  values are 'columns' to shrink the range at its left and right
         *  border, or 'rows' to shrink the range at its top and bottom border.
         *
         * @returns {Object|Null}
         *  The address of the shrunk range; or null, if no valid range was
         *  left after shrinking (the range has been shrunk to zero
         *  columns/rows).
         */
        this.shrinkRangeFromMergedRanges = function (range, direction) {

            var // flat copy of the merged ranges, to prevent processing them multiple times
                pendingRanges = _.clone(mergedRanges),
                // the length of the pending ranges array
                pendingLength = pendingRanges.length,
                // address array index for passed direction
                addressIndex = (direction === 'columns') ? 0 : 1;

            // loop iterator function must be defined outside the do-while loop
            function shrinkFromMergedRange(mergedRange, index) {

                // do not process merged ranges that are fully contained in the current range
                if (SheetUtils.rangeContainsRange(range, mergedRange)) { return; }
                // remove the merged range from the pending array
                pendingRanges.splice(index, 1);
                // nothing to do, if merged range is completely outside
                if (!SheetUtils.rangeOverlapsRange(range, mergedRange)) { return; }

                // shrink leading border of the range
                if ((mergedRange.start[addressIndex] < range.start[addressIndex]) && (range.start[addressIndex] <= mergedRange.end[addressIndex])) {
                    range.start[addressIndex] = mergedRange.end[addressIndex] + 1;
                }
                // shrink trailing border of the range
                if ((mergedRange.start[addressIndex] <= range.end[addressIndex]) && (range.end[addressIndex] < mergedRange.end[addressIndex])) {
                    range.end[addressIndex] = mergedRange.start[addressIndex] - 1;
                }

                // check if the range has been shrunk to null size
                if (range.start[addressIndex] > range.end[addressIndex]) {
                    range = null;
                    return Utils.BREAK;
                }
            }

            // loop as long as the range will still be shrunk to exclude other merged ranges
            if (pendingRanges.length > 0) {
                do {
                    // store the current length of the pending ranges array (checking length afterwards)
                    pendingLength = pendingRanges.length;
                    // iterate backwards to be able to remove an element from the array
                    Utils.iterateArray(pendingRanges, shrinkFromMergedRange, { reverse: true });
                } while (range && (pendingRanges.length < pendingLength));
            }

            // return the shrunk range
            return range;
        };

        /**
         * Merges the specified cell range, or removes any merged ranges from
         * that range.
         *
         * @param {Object} range
         *  The address of the range to be merged or unmerged.
         *
         * @param {String} type
         *  The merge type. Must be one of: 'merge', 'horizontal', 'vertical',
         *  or 'unmerge'.
         *
         * @returns {MergeCollection}
         *  A reference to this instance.
         */
        this.mergeRange = function (range, type) {

            var // all existing merged ranges deleted from the collection
                deletedRanges = [],
                // all merged ranges added to the collection
                insertedRanges = [];

            // inserts the passed range into the collection, and into the 'insertedRanges' array
            function insertMergedRange(mergedRange) {

                var // the array index of the passed range in the 'deletedRanges' array
                    deletedIndex = Utils.findFirstIndex(deletedRanges, _.partial(_.isEqual, mergedRange));

                // if the range has been deleted, remove it from the array, otherwise
                // insert it into the 'insertedRanges' array (this prevents to notify
                // ranges that have been unmerged *and* merged)
                if (deletedIndex >= 0) {
                    deletedRanges.splice(deletedIndex, 1);
                } else {
                    insertedRanges.push(mergedRange);
                }

                // push a copy of the range into the collection
                mergedRanges.push(_.copy(mergedRange, true));
            }

            // always remove all merged ranges that overlap with the passed range,
            // regardless of passed type (needed to merge over merged ranges)
            mergedRanges = _.reject(mergedRanges, function (mergedRange) {
                var overlaps = SheetUtils.rangeOverlapsRange(range, mergedRange);
                if (overlaps) { deletedRanges.push(mergedRange); }
                return overlaps;
            });

            // handle the different merge types
            switch (type) {
            case 'merge':
                if (SheetUtils.getCellCount(range) > 1) {
                    insertMergedRange(range);
                }
                break;
            case 'horizontal':
                if (SheetUtils.getColCount(range) > 1) {
                    Utils.iterateRange(range.start[1], range.end[1] + 1, function (row) {
                        insertMergedRange({ start: [range.start[0], row], end: [range.end[0], row] });
                    });
                }
                break;
            case 'vertical':
                if (SheetUtils.getRowCount(range) > 1) {
                    Utils.iterateRange(range.start[0], range.end[0] + 1, function (col) {
                        insertMergedRange({ start: [col, range.start[1]], end: [col, range.end[1]] });
                    });
                }
                break;
            case 'unmerge':
                // all merged ranges have been removed already, see above
                break;
            default:
                Utils.warn('MergeCollection.mergeRange(): unknown merge type "' + type + '"');
            }

            // trigger change events with array of changed merged ranges
            triggerEvent('delete:merged', deletedRanges);
            triggerEvent('insert:merged', insertedRanges);

            return this;
        };

        /**
         * Transforms the passed merged range according to the specified column
         * or row operation. Does NOT modify any merged range in this merged
         * ranges collection.
         *
         * @param {Object} mergedRange
         *  The address of the merged range to be transformed.
         *
         * @param {Object} operationInterval
         *  The column/row interval of the operation, with the index properties
         *  'first' and 'last'.
         *
         * @param {Boolean} insert
         *  Whether the specified column/row interval has been inserted into
         *  the sheet (true), or deleted from the sheet (false).
         *
         * @param {Boolean} columns
         *  Whether the specified interval is a column interval (true), or a
         *  row interval (false).
         *
         * @returns {Object|Null}
         *  The address of the transformed merged range. If the range has been
         *  deleted completely while deleting columns/rows, or has been shifted
         *  outside the sheet completely while inserting columns/rows, or if
         *  the resulting merged range covers a single cell only, null will be
         *  returned instead.
         */
        this.transformRange = function (mergedRange, operationInterval, insert, columns) {
            // transform the passed range without expanding the end of the range while inserting
            mergedRange = docModel.transformRange(mergedRange, operationInterval, insert, columns);
            // a transformed merged range covering a single cell is considered deleted
            return (mergedRange && (SheetUtils.getCellCount(mergedRange) > 1)) ? mergedRange : null;
        };

        // initialization -----------------------------------------------------

        // update merged ranges after inserting/deleting columns or rows in the sheet
        sheetModel.registerTransformationHandler(transformationHandler);

        // clone merged ranges passed as hidden argument by the clone() method
        if (_.isArray(arguments[MergeCollection.length])) {
            mergedRanges = _.copy(arguments[MergeCollection.length], true);
        }

        // destroy all class members on destruction
        this.registerDestructor(function () {
            app = sheetModel = self = docModel = null;
        });

    } // class MergeCollection

    // exports ================================================================

    // derive this class from class ModelObject
    return ModelObject.extend({ constructor: MergeCollection });

});
