/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2016 OX Software GmbH.
 *
 * @author Michael Nimz <michael.nimz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/view/popup/sortlayermenu', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/control/checkbox',
    'io.ox/office/tk/control/radiogroup',
    'io.ox/office/tk/control/radiolist',
    'io.ox/office/baseframework/view/popup/layermenu',
    'io.ox/office/spreadsheet/utils/sheetutils',
    'io.ox/office/spreadsheet/view/labels',
    'gettext!io.ox/office/spreadsheet'
], function (Utils, CheckBox, RadioGroup, RadioList, LayerMenu, SheetUtils, Labels, gt) {

    'use strict';

    // class SortLayerMenu ====================================================

    /**
     * A layer menu with options for sorting.
     *
     * @constructor
     *
     * @extends LayerMenu
     */
    function SortLayerMenu(app) {

        var // self reference
            self = this,

            // the document view
            docView = null,

            directionGroup = null,
            keyRadioList = null,
            currentDirection = 'vertical',
            maxCount = 2000;

        // base constructor ---------------------------------------------------

        LayerMenu.call(this, app, /*#. menu title: custom sort options */ gt.pgettext('sort', 'Custom Sort'), {
            actionKey: 'cell/sort/custom',
            actionValue: 'custom'
        });

        // private methods ----------------------------------------------------

        function fillRadioList(columns) {

            var currentSelection = docView.getSelection(),
                interval = SheetUtils.getInterval(currentSelection.ranges[0], columns),
                from = interval.first,
                to = interval.last,
                controllerDefault = null;

            // remove (old) column/row values. New selection will have new values
            keyRadioList.getMenu().clearContents();

            if (SheetUtils.getIntervalSize(interval) > maxCount) { return false; }

            for (var index = from; index <= to; index += 1) {
                // set default selection to the first entry
                if (_.isNull(controllerDefault)) { controllerDefault = index; }

                keyRadioList.createOptionButton(index, { label: columns ? Labels.getColLabel(index) : Labels.getRowLabel(index) });
            }
            app.getController().executeItem('cell/sort/colrow', controllerDefault);
        }

        /**
         * Initializes the columns/rows radio list, after the sorting direction
         * has been changed.
         */
        function changeDirectionHandler(event, val) {
            if (currentDirection !== val) {
                currentDirection = val;
                fillRadioList(val === 'vertical');
            }
        }

        /**
         * Closes the menu when the selection in the document has been changed.
         */
        function changeSelectionHandler() {
            if (docView.getSortOption('preventClose', 'custom')) {
                docView.setSortOption('preventClose', false, 'custom');
            } else {
                self.hide();
            }
        }

        /**
         * Fills the radio lists before the menu will be shown.
         */
        function popupBeforeShowHandler() {
            fillRadioList(true);
        }

        /**
         * Resets sort options after the menu has been closed.
         */
        function popupHideHandler() {
            docView.resetSortOptions('custom');
        }

        /**
         * Moves the menu to a default position when it has been shown for the
         * first time.
         */
        function setDefaultPosition() {
            var pos = Utils.getNodePositionInPage(app.getView().getAppPaneNode());
            self.setNodePosition(pos.left + (pos.width - self.getNode().outerWidth()) / 2, pos.top + 40);
        }

        // initialization -----------------------------------------------------

        // initialize after application is ready
        app.onInit(function () {
            docView = app.getView();

            // add base menu items
            self.addGroup('cell/sort/hasheadline', new CheckBox({ label: /*#. sorting: selected range has headers which shouldn't be sorted */ gt.pgettext('sort', 'Selection has headers'), boxed: true }))
                .addSectionLabel(/*#. sorting: the sort direction headline for 'top to bottom' or 'left to right' */ gt.pgettext('sort', 'Direction'))
                .addGroup('cell/sort/direction', directionGroup = new RadioGroup({ label: gt.pgettext('sort', 'Direction'), radioIcons: true })
                    .createOptionButton('vertical', { label: /*#. sort direction 'Top to bottom' */ gt.pgettext('sort', 'Top to bottom') })
                    .createOptionButton('horizontal', { label: /*#. sort direction 'Left to right' */ gt.pgettext('sort', 'Left to right') }),
                { inline: true })
                .addSectionLabel( /*#. head line: 'sort by' options (list of column names, e.g. 'Column C') */ gt.pgettext('sort', 'Sort by'))
                .addGroup('cell/sort/colrow', keyRadioList = new RadioList({ style: 'width: 150px;', updateCaptionMode: 'label' }), { inline: true })
                .addGroup('cell/sort/order', new RadioList({ style: 'width:150px;text-align:right;' })
                    .createOptionButton('ascending', { label: /*#. sort ascending */ gt.pgettext('sort', 'Ascending') })
                    .createOptionButton('descending', { label: /*#. sort descending */ gt.pgettext('sort', 'Descending') }),
                { inline: true });

            // refill the column/row selector on changing the direction
            self.listenTo(directionGroup, 'group:change', changeDirectionHandler);

            // reset the custom option 'preventClose' on 'change:selection'
            // event and do nothing, otherwise hide the pop-up menu
            self.listenTo(docView, 'change:selection', changeSelectionHandler);

            // add event handlers for the menu
            self.on({ 'popup:beforeshow': popupBeforeShowHandler, 'popup:hide': popupHideHandler });

            // set default position on first display of the menu
            self.one('popup:show', setDefaultPosition);
        });

        // destroy all class members
        this.registerDestructor(function () {
            app = self = docView = null;
        });

    } // class SortLayerMenu

    // exports ================================================================

    // derive this class from class LayerMenu
    return LayerMenu.extend({ constructor: SortLayerMenu });

});
