/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2016 OX Software GmbH.
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/text/app/controller',
    ['io.ox/office/tk/utils',
     'io.ox/office/editframework/utils/border',
     'io.ox/office/editframework/utils/mixedborder',
     'io.ox/office/editframework/utils/lineheight',
     'io.ox/office/editframework/app/editcontroller',
     'io.ox/office/drawinglayer/view/drawinglabels',
     'io.ox/office/drawinglayer/view/imageutil',
     'io.ox/office/text/utils/config',
     'io.ox/office/text/hyperlink',
     'io.ox/office/text/position',
     'io.ox/office/text/format/listcollection',
     'io.ox/office/text/format/paragraphstyles',
     'io.ox/office/text/format/tablestyles',
     'io.ox/office/text/format/drawingstyles'
    ], function (Utils, Border, MixedBorder, LineHeight, EditController, DrawingLabels, Image, Config, Hyperlink, Position, ListCollection, ParagraphStyles, TableStyles, DrawingStyles) {

    'use strict';

    // class TextController ===================================================

    /**
     * The controller of a OX Text application.
     *
     * @constructor
     *
     * @extends EditController
     */
    function TextController(app) {

        var // self reference
            self = this,

            // the text model of the passed application
            model = null,

            // the view instance
            view = null,

            // all the little controller items
            items = {

                // view -------------------------------------------------------

                'view/zoom': {
                    parent: 'app/valid',
                    get: function () { return view.getZoomFactor() / 100; },
                    set: function (zoom) { view.setZoomType(zoom * 100); }
                },

                'view/zoom/dec': {
                    parent: 'app/valid',
                    enable: function () { return view.getZoomFactor() > view.getMinZoomFactor(); },
                    set: function () { view.decreaseZoomLevel(); }
                },

                'view/zoom/inc': {
                    parent: 'app/valid',
                    enable: function () { return view.getZoomFactor() < view.getMaxZoomFactor(); },
                    set: function () { view.increaseZoomLevel(); }
                },

                'view/zoom/type': {
                    parent: 'app/valid',
                    get: function () { return view.getZoomType(); },
                    set: function (zoomType) { view.setZoomType(zoomType); }
                },

                'document/editable/text': {
                    parent: 'document/editable',
                    enable: function () { return model.isTextSelected(); }
                },

                'document/editable/text/notextframe': {
                    parent: 'document/editable/text',
                    enable: function () {
                        return !model.getSelection().isAdditionalTextframeSelection();
                    }
                },

                'document/selectall': {
                    // enabled in read-only mode
                    set: function () { model.selectAll(); },
                    // restrict keyboard shortcut to application pane
                    shortcut: { keyCode: 'A', ctrlOrMeta: true, selector: '.app-pane' },
                    preserveFocus: true // do not return focus to document
                },

                'document/selectDrawing': {
                    set: function (options) {
                        model.getSelection().selectNextDrawing({backwards: Utils.getBooleanOption(options, 'backwards', false)});
                    },
                    shortcut: { keyCode: 'F4', shift: true }
                },
                'document/setCursorIntoTextframe': {
                    set: function () {
                        model.getSelection().setSelectionIntoTextframe();
                    },
                    shortcut: { keyCode: 'F2' }
                },

                // document contents

                'document/search/highlight': {
                    // enabled in read-only mode
                    set: function (query) { model.quickSearch(query, 0); },
                    preserveFocus: true // do not return focus to document
                },
                'document/search/nextResult': {
                    // enabled in read-only mode
                    set: function () { model.selectNextSearchResult(); },
                    shortcut: [{ keyCode: 'G', ctrlOrMeta: true }, { keyCode: 'F3' }],
                    preserveFocus: true // do not return focus to document
                },
                'document/search/previousResult': {
                    // enabled in read-only mode
                    set: function () { model.selectPreviousSearchResult(); },
                    shortcut: [{ keyCode: 'G', ctrlOrMeta: true, shift: true }, { keyCode: 'F3', shift: true }],
                    preserveFocus: true // do not return focus to document
                },
                'document/search/replaceAll': {
                    parent: 'document/editable',
                    set: function () { model.searchAndReplaceAll(app.getWindow().search.query, app.getWindow().search.replacement); }
                },
                'document/search/replaceSelected': {
                    parent: 'document/editable',
                    set: function () { model.replaceSelectedSearchResult(app.getWindow().search.replacement); }
                },

                'document/cut': {
                    parent: 'document/editable',
                    enable: function () { return model.hasSelectedRange(); },
                    set: function () { model.cut(); }
                },
                'document/copy': {
                    // enabled in read-only mode
                    enable: function () { return model.hasSelectedRange(); },
                    set: function () { model.copy(); }
                },
                'document/paste': {
                    parent: 'document/editable',
                    enable: function () { return model.hasInternalClipboard(); },
                    set: function () { model.pasteInternalClipboard(); }
                },

                // spelling

                'document/spelling/available': {
                    parent: 'document/editable',
                    enable: function () { return Config.SPELLING_ENABLED; }
                },
                'document/onlinespelling': {
                    parent: 'document/spelling/available',
                    get: function () { return model.isOnlineSpelling(); },
                    set: function (state) { model.setOnlineSpelling(state); }
                },

                // paragraphs

                'paragraph/group': {
                    parent: 'document/editable/text',
                    get: function () { return model.getAttributes('paragraph', {maxIterations: 10}) || {}; }
                },
                'paragraph/stylesheet': {
                    parent: 'paragraph/group',
                    get: function (paragraph) { return paragraph.styleId; },
                    set: function (styleId) { model.setAttributes('paragraph', { styleId: styleId }, { clear: true }); }
                },
                'paragraph/attributes': {
                    parent: 'paragraph/group',
                    get: function (paragraph) { return paragraph.paragraph || {}; }
                },
                'paragraph/alignment': {
                    parent: 'paragraph/attributes',
                    get: function (attributes) { return attributes.alignment; },
                    set: function (alignment) { model.setAttribute('paragraph', 'alignment', alignment); }
//                  removed for the 7.2.0 due to clashes with browser shortcuts
//                    shortcut: [
//                        { keyCode: 'L', ctrlOrMeta: true, value: 'left' },
//                        { keyCode: 'R', ctrlOrMeta: true, value: 'right' },
//                        { keyCode: 'E', ctrlOrMeta: true, value: 'center' },
//                        { keyCode: 'J', ctrlOrMeta: true, value: 'justify' }
//                    ]
                },
                'paragraph/lineheight': {
                    parent: 'paragraph/attributes',
                    get: function (attributes) { return attributes.lineHeight; },
                    set: function (lineHeight) { model.setAttribute('paragraph', 'lineHeight', lineHeight); }
//                  removed for the 7.2.0 due to clashes with browser shortcuts
//                    shortcut: [
//                        { keyCode: '1', ctrlOrMeta: true, value: LineHeight.SINGLE },
//                        { keyCode: '5', ctrlOrMeta: true, value: LineHeight.ONE_HALF },
//                        { keyCode: '2', ctrlOrMeta: true, value: LineHeight.DOUBLE }
//                    ]
                },
                'paragraph/fillcolor': {
                    parent: 'paragraph/attributes',
                    get: function (attributes) { return attributes.fillColor; },
                    set: function (color) { model.setAttribute('paragraph', 'fillColor', color); }
                },
                'paragraph/borders': {
                    parent: 'paragraph/attributes',
                    get: function (attributes) { return MixedBorder.getBorderMode(attributes, { paragraph: true }); },
                    set: function (borderMode) {
                        var borderAttrs = MixedBorder.getBorderAttributes(borderMode, this.getParentValue(), ParagraphStyles.SINGLE_BORDER, { paragraph: true });
                        model.setAttributes('paragraph', { paragraph: borderAttrs });
                    }
                },

                // toggle default bullet list, or select bullet type
                'paragraph/list/bullet': {
                    parent: 'paragraph/attributes',
                    get: _.partial(getListStyleId, 'bullet'),
                    set: _.partial(setListStyleId, 'bullet')
                },
                // toggle default numbered list, or select numbering type
                'paragraph/list/numbered': {
                    parent: 'paragraph/attributes',
                    get: _.partial(getListStyleId, 'numbering'),
                    set: _.partial(setListStyleId, 'numbering')
                },

                // parent for controller items only enabled inside lists
                'paragraph/list/enabled': {
                    parent: 'paragraph/attributes',
                    enable: function () {
                        var indent = this.getValue().listLevel;
                        return _.isNumber(indent) && (0 <= indent) && (indent <= 8);
                    }
                },
                // change list level
                'paragraph/list/indent': {
                    parent: 'paragraph/list/enabled',
                    get: function (attributes) { return attributes.listLevel; },
                    set: function (indent) { model.setAttribute('paragraph', 'listLevel', indent); }
                },
                // increase list level by one
                'paragraph/list/incindent': {
                    parent: 'paragraph/list/indent',
                    enable: function () { return this.getValue() < 8; },
                    set: function () {
                        var indent = this.getValue();
                        if (indent < 8) {
                            model.setAttribute('paragraph', 'listLevel', indent + 1);
                        }
                    }
                },
                // decrease list level by one
                'paragraph/list/decindent': {
                    parent: 'paragraph/list/indent',
                    enable: function () { return this.getValue() > 0; },
                    set: function () {
                        var indent = this.getValue();
                        if (indent > 0) {
                            model.setAttribute('paragraph', 'listLevel', indent - 1);
                        }
                    }
                },

                // characters

                'character/group': {
                    parent: 'document/editable/text',
                    get: function () { return model.getAttributes('character', {maxIterations: 10}) || {}; }
                },
                'character/stylesheet': {
                    parent: 'character/group',
                    get: function (character) { return character.styleId; },
                    set: function (styleId) { model.setAttributes('character', { styleId: styleId }, { clear: true }); }
                },
                'character/attributes': {
                    parent: 'character/group',
                    get: function (character) { return character.character || {}; }
                },
                'character/fontname': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.fontName; },
                    set: function (fontName) { model.setAttribute('character', 'fontName', fontName); }
                },
                'character/fontsize': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.fontSize; },
                    set: function (fontSize) { model.setAttribute('character', 'fontSize', fontSize); }
                },
                'character/bold': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.bold; },
                    set: function (state) { model.setAttribute('character', 'bold', state); },
                    shortcut: { keyCode: 'B', ctrlOrMeta: true, value: function (state) { return !state; } }
                },
                'character/italic': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.italic; },
                    set: function (state) { model.setAttribute('character', 'italic', state); },
                    shortcut: { keyCode: 'I', ctrlOrMeta: true, value: function (state) { return !state; } }
                },
                'character/underline': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.underline; },
                    set: function (state) { model.setAttribute('character', 'underline', state); },
                    shortcut: { keyCode: 'U', ctrlOrMeta: true, value: function (state) { return !state; } }
                },
                'character/strike': {
                    parent: 'character/attributes',
                    get: function (attributes) { return _.isString(attributes.strike) ? (attributes.strike !== 'none') : null; },
                    set: function (state) { model.setAttribute('character', 'strike', state ? 'single' : 'none'); }
                },
                'character/vertalign': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.vertAlign; },
                    set: function (align) { model.setAttribute('character', 'vertAlign', align); }
                },
                'character/color': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.color; },
                    set: function (color) { model.setAttribute('character', 'color', color); }
                },
                'character/fillcolor': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.fillColor; },
                    set: function (color) { model.setAttribute('character', 'fillColor', color); }
                },
                'character/language': {
                    parent: 'character/attributes',
                    get: function (attributes) { return attributes.language; },
                    set: function (language) { model.setAttribute('character', 'language', language); }
                },
                'character/hyperlink': {
                    parent: 'character/attributes',
                    enable: function () { return model.hasEnclosingParagraph(); },
                    get: function (attributes) { return attributes.url; }
                    // no direct setter (see item 'character/hyperlink/dialog')
                },
                'character/hyperlink/dialog': {
                    parent: 'character/hyperlink',
                    get: function () { return Hyperlink.hasPopup(model); },
                    set: function () { return model.insertHyperlinkDialog(); }
                    // removed for the 7.2.0 due to clashes with browser shortcuts
//                    shortcut: { keyCode: 'K', ctrlOrMeta: true }
                },
                'character/hyperlink/remove': {
                    parent: 'character/hyperlink',
                    set: function () { return model.removeHyperlink(); }
                },

                'character/reset': {
                    parent: 'document/editable/text',
                    set: function () { model.resetAttributes(); },
                    shortcut: { keyCode: 'SPACE', ctrl: true }
                },
                'character/insert/break': {
                    parent: 'document/editable/text',
                    set: function () { return model.insertHardBreak(); }
                },
                'character/insert/tab': {
                    parent: 'document/editable/text',
                    set: function () { return model.insertTab(); }
                },
                'character/insert/pagebreak': {
                    parent: 'document/editable/text/notextframe',
                    set: function () { return model.insertManualPageBreak(); }
                },

                // reduced odf text frame functionality
                'insert/textframe/supported': {
                    enable: function () { return !model.isReducedOdfTextframeFunctionality(); }
                },

                // tables

                'table/group': {
                    parent: 'document/editable/text',
                    get: function () { return model.getAttributes('table') || {}; }
                },
                'table/insert/available': {
                    enable: function () { return Config.MAX_TABLE_CELLS > 0; }
                },
                'table/insert': {
                    parent: ['document/editable/text', 'table/insert/available', 'insert/textframe/supported'],
                    set: function (size) { model.insertTable(size); },
                    preserveFocus: true // do not return focus to document
                },

                'document/editable/table': {
                    parent: 'table/group',
                    enable: function () { return model.isPositionInTable(); }
                },
                'table/insert/row': {
                    parent: 'document/editable/table',
                    set: function () { model.insertRow(); }
                },
                'table/insert/column': {
                    parent: 'document/editable/table',
                    set: function () { model.insertColumn(); }
                },
                'table/delete/row': {
                    parent: 'document/editable/table',
                    set: function () { return model.deleteRows(); }
                },
                'table/delete/column': {
                    parent: 'document/editable/table',
                    set: function () { return model.deleteColumns(); }
                },

                'table/stylesheet': {
                    parent: 'document/editable/table',
                    get: function (table) { return table.styleId; },
                    set: function (styleId) { model.setAttributes('table', { styleId: styleId }, { clear: true }); }
                },
                'table/attributes': {
                    parent: 'document/editable/table',
                    get: function (table) { return table.table || {}; }
                },
                'table/cellattributes': {
                    parent: 'document/editable/table',
                    get: function () { return model.getAttributes('cell').cell || {}; }
                },
                'table/cellborder': {
                    parent: 'table/attributes',
                    get: function (attributes) { return MixedBorder.getBorderMode(attributes); },
                    set: function (borderMode) {
                        var borderAttrs = MixedBorder.getBorderAttributes(borderMode, this.getParentValue(), ParagraphStyles.SINGLE_BORDER);
                        model.setAttributes('table', { table: borderAttrs }, { cellSpecificTableAttribute: true });
                    }
                },
                'table/borderwidth': {
                    parent: 'table/cellattributes',
                    get: function (attributes) { return TableStyles.getBorderStyleFromAttributes(attributes); },
                    set: function (borderWidth) {
                        model.setAttributes('table', {
                            table: TableStyles.getAttributesFromBorderStyle(borderWidth, model.getAttributes('table').table)
                        }, { onlyVisibleBorders: true });
                    }
                },
                'table/fillcolor': {
                    parent: 'table/cellattributes',
                    get: function (attributes) { return attributes.fillColor; },
                    set: function (color) { model.setAttribute('cell', 'fillColor', color); }
                },
                'table/split': {
                    parent: 'document/editable/table',
                    set: function () { model.splitTable(); }
                },

                // drawing

                'document/editable/drawing': {
                    parent: 'document/editable',
                    enable: function () { return model.isDrawingSelected(); }
                },
                'document/editable/anydrawing': {
                    parent: 'document/editable',
                    enable: function () { return model.getSelection().isAnyDrawingSelection(); }
                },
                'document/editable/onlysupporteddrawings': {
                    parent: 'document/editable',
                    enable: function () { return model.getSelection().areOnlySupportedDrawingsSelected(); }
                },
                'drawing/shapes/support/fill': {
                    parent: 'document/editable/onlysupporteddrawings',
                    enable: function () { return (/^(shape|group)$/).test(model.getSelection().getSelectedDrawingType()) || (model.getSelection().getSelectionType() === 'text'); }
                },
                'drawing/shapes/support/line': {
                    parent: 'document/editable/onlysupporteddrawings',
                    enable: function () { return (/^(shape|group|image)$/).test(model.getSelection().getSelectedDrawingType()) || (model.getSelection().getSelectionType() === 'text'); }
                },

                'drawing/attributes': {
                    parent: 'document/editable/anydrawing',
                    get: function () { return model.getAttributes('drawing').drawing || {}; }
                },
                'drawing/group/attributes': {
                    parent: 'document/editable/anydrawing',
                    get: function () { return model.getAttributes('drawing', {groupOnly: true}).drawing || {}; }
                },

                'drawing/delete': {
                    parent: 'document/editable/anydrawing',
                    set: function () {
                        if (model.isDrawingSelected()) {
                            return model.deleteSelected();

                        } else if (model.getSelection().isAdditionalTextframeSelection()) {
                            var start = Position.getOxoPosition(model.getNode(), model.getSelection().getSelectedTextFrameDrawing(), 0),
                                end = Position.increaseLastIndex(start);

                            model.getSelection().setTextSelection(start, end);
                            return model.deleteSelected();
                        }
                    }
                },
                'drawing/anchorTo' : {
                    parent: 'drawing/group/attributes',
                    enable: function () { return !model.getSelection().isDrawingInTextFrameSelection() || model.getSelection().isDrawingInODFTextFrameSelection(); },
                    get: function (attributes) { return (attributes.inline) ? 'inline' : attributes.anchorVertBase; },
                    set: function (value) { model.anchorDrawingTo(value); }
                },
                'drawing/position': {
                    parent: 'drawing/group/attributes',
                    enable: function (attributes) { return (!attributes.inline); },
                    get: function (attributes) { return DrawingStyles.getPositionFromAttributes(attributes); },
                    set: function (position) {
                        model.setAttributes('drawing', {
                            drawing: DrawingStyles.getAttributesFromPosition(position, model.getAttributes('drawing').drawing)
                        }, {
                            ctSupportsDrawingPosition: model.getChangeTrack().ctSupportsDrawingPosition()
                        });
                    }
                },

                'drawing/type/label': {
                    parent: 'document/editable/drawing',
                    get: function () { return DrawingLabels.getDrawingTypeLabel(model.getSelection().getClosestSelectedDrawingType()); }
                },

                // drawing fill formatting

                'drawing/fill/attributes': {
                    parent: 'document/editable/anydrawing',
                    get: function () { return model.getAttributes('drawing').fill || {}; }
                },

                'drawing/fill/color' : {
                    parent: 'drawing/fill/attributes',
                    get: function (attributes) { return attributes.color; },
                    set: function (color) { model.setDrawingFillColor(color); }
                },

                // drawing line formatting

                'drawing/line/attributes': {
                    parent: 'document/editable/anydrawing',
                    get: function () { return model.getAttributes('drawing').line || {}; }
                },

                'drawing/line/style' : {
                    parent: 'drawing/line/attributes',
                    get: function (attributes) { return model.detectBorderStyle(attributes); },
                    set: function (border) { model.setDrawingBorder(border); }
                },

                'drawing/line/color' : {
                    parent: 'drawing/line/attributes',
                    get: function (attributes) { return (attributes.type !== 'none') ? attributes.color : 'auto'; },
                    set: function (color) { model.setDrawingBorderColor(color); }
                },

                // text frames

                'textframe/insert' : {
                    parent: 'document/editable/text/notextframe',
                    set: function () { model.insertTextFrame(); }
                },

                // images

                'image/insert/dialog': {
                    parent: ['document/editable/text', 'insert/textframe/supported'],
                    set: function () { return model.showInsertImageDialog(); }
                },

                // change tracking

                'checkCurrentChangeTracking' : {
                    parent: 'document/editable',
                    enable: function(){ return model.getChangeTrack().isOnChangeTrackNode(); }
                },
                'toggleChangeTracking' : {
                    parent: 'document/editable',
                    get: function () { return model.getChangeTrack().isActiveChangeTracking(); },
                    set: function (state) { model.getChangeTrack().setActiveChangeTracking(state); }
                },
                'acceptChangeTracking' : {
                    parent: 'document/editable',
                    set: function () {
                        model.getChangeTrack().resolveChangeTracking(null, { accepted: true, all: true });
                        view.getChangeTrackPopup().hide();
                    }
                },
                'rejectChangeTracking' : {
                    parent: 'document/editable',
                    set: function () {
                        model.getChangeTrack().resolveChangeTracking(null, { accepted: false, all: true });
                        view.getChangeTrackPopup().hide();
                    }
                },
                'acceptSelectedChangeTracking' : {
                    parent: 'checkCurrentChangeTracking',
                    set: function () {
                        model.getChangeTrack().resolveChangeTracking(null, { accepted: true, all: false });
                        view.getChangeTrackPopup().hide();
                    }
                },
                'acceptMoveSelectedChangeTracking' : {
                    parent: 'document/editable',
                    set: function () {
                        model.getChangeTrack().resolveChangeTracking(null, { accepted: true, all: false });
                        view.getChangeTrackPopup().hide();
                        model.getChangeTrack().getNextChangeTrack();
                    }
                },
                'rejectSelectedChangeTracking' : {
                    parent: 'checkCurrentChangeTracking',
                    set: function () {
                        model.getChangeTrack().resolveChangeTracking(null, { accepted: false, all: false });
                        view.getChangeTrackPopup().hide();
                    }
                },
                'rejectMoveSelectedChangeTracking' : {
                    parent: 'document/editable',
                    set: function () {
                        model.getChangeTrack().resolveChangeTracking(null, { accepted: false, all: false });
                        view.getChangeTrackPopup().hide();
                        model.getChangeTrack().getNextChangeTrack();
                    }
                },
                'selectNextChangeTracking' : {
                    parent: 'document/editable',
                    set: function () { model.getChangeTrack().getNextChangeTrack({ next: true }); }
                },
                'selectPrevChangeTracking' : {
                    parent: 'document/editable',
                    set: function () { model.getChangeTrack().getNextChangeTrack({ next: false }); }
                },
                'changetrackPopupShow' : {
                    parent: 'document/editable/text',
                    set: function () { view.changeTrackPopupSelect(); }
                },

                // debug
                'format/spellchecking' : {
                    parent: 'character/attributes',
                    enable: function () { return model.hasSelectedRange() && model.hasEnclosingParagraph(); },
                    set: function () { model.checkSpelling(); }
                },
                'debug/accessrights' : {
                    parent: 'document/editable/text',
                    enable: function () { return model.hasSelectedRange(); },
                    set: function (mail) { model.accessRightsExport(mail); }
                },
                'debug/insertDrawingGroup' : {
                    set: function () { model.insertDrawingGroup(); }
                },
                'debug/uselocalstorage' : {
                    get: function () { return app.getUserSettingsValue('debugUseLocalStorage', Config.USE_LOCAL_STORAGE); },
                    set: function (state) { app.setUserSettingsValue('debugUseLocalStorage', state); }
                },
                'debug/useFastLoad' : {
                    get: function () { return app.getUserSettingsValue('documentFastLoad', true); },
                    set: function (state) { app.setUserSettingsValue('documentFastLoad', state); }
                },
                'debug/recordoperations' : {
                    get: function () { return model.isRecordingOperations(); },
                    set: function (state) { model.setRecordingOperations(state); }
                },
                'debug/replayoperations' : {
                    parent: 'document/editable',
                    enable: function () { return app.getView().getDebugOperationsContent() !== null; },
                    set: function () { model.replayOperations(app.getView().getDebugOperationsContent()); }
                },
                'debug/pagebreaks/toggle' : {
                    parent: 'debug/enabled',
                    enable: function () { return !model.isDraftMode() && !Utils.SMALL_DEVICE; },
                    get: function () { return model.isPageBreakMode(); },
                    set: function (state) { model.togglePageBreakMode(state); }
                },
                'debug/draftmode/toggle' : {
                    parent: 'debug/enabled',
                    enable: function () { return !Utils.SMALL_DEVICE; },
                    get: function () { return model.isDraftMode(); },
                    set: function (state) { model.toggleDraftMode(state); }
                },

                // android-entry for selection which cant be deleted by softkeyboard
                'document/editable/selection': {
                    parent: 'document/editable',
                    enable: function () {
                        var selection = model.getSelection();
                        return !_.isEqual(selection.getStartPosition(), selection.getEndPosition());
                    }
                },
                'selection/delete': {
                    parent: 'document/editable/selection',
                    set: function () { return model.deleteSelected(); }
                },

            };

        // base constructor ---------------------------------------------------

        EditController.call(this, app, { updateDelay: 100 });

        // private methods ----------------------------------------------------

        /**
         * Item getter function for a list style identifier. Will be called in
         * the context of a controller item instance.
         *
         * @param {String} listType
         *  The type of the list styles: either 'bullet' or 'numbering'.
         *
         * @param {Object} attributes
         *  The merged attribute set of the selected paragraphs.
         *
         * @returns {String|Null}
         *  If unambiguous, the list style identifier contained in the passed
         *  paragraph attributes, if the type of that list matches the list
         *  type specified in the item user data ('bullets' or 'numbering'),
         *  otherwise an empty string. If ambiguous, returns the value null.
         */
        function getListStyleId(listType, attributes) {

            var listStyleId = attributes.listStyleId,
                listLevel = attributes.listLevel,
                isListType = false;

            // ambiguous attributes
            if (_.isNull(listStyleId) || _.isNull(listLevel)) { return null; }

            if (listLevel > -1) {
                switch (listType) {
                case 'bullet':
                    isListType = model.getListCollection().isBulletsList(listStyleId, listLevel);
                    break;
                case 'numbering':
                    isListType = model.getListCollection().isNumberingList(listStyleId, listLevel);
                    break;
                }
            }

            // return empty string, if list type in attributes does not match list type in item user data
            return isListType ? listStyleId : '';
        }

        /**
         * Item setter function for a list style identifier. Will be called in
         * the context of a controller item instance.
         *
         * @param {String} listType
         *  The type of the list styles: either 'bullet' or 'numbering'.
         *
         * @param {String} listStyleId
         *  The new list style identifier. The special value
         *  Lists.DEFAULT_VALUE can be used to toggle the default list style.
         */
        function setListStyleId(listType, listStyleId) {

            // simulate toggle behavior for default list style
            if (listStyleId === ListCollection.DEFAULT_VALUE) {
                if (this.getValue() === '') {
                    model.createDefaultList(listType);
                } else {
                    model.removeListAttributes();
                }
            } else {
                // list level may be null, will fall back to level 0 then...
                model.createSelectedListStyle(listStyleId, self.getItemValue('paragraph/list/indent'));
            }
        }

        // initialization -----------------------------------------------------

        if(_.browser.Android){
            var originalUpdate = self.update;

            self.update = function() {
                if (model.isAndroidTyping()) {
                    self.executeDelayed(self.update, 1000);
                } else {
                    originalUpdate();
                }
            };
        }

        // register item definitions
        this.registerDefinitions(items);

        // initialization after construction
        app.onInit(function () {
            // model and view may not be available at construction time
            model = app.getModel();
            view = app.getView();
            // update GUI after changed selection
            model.on('selection', function () { self.update(); });
        });

    } // class TextController

    // exports ================================================================

    // derive this class from class EditController
    return EditController.extend({ constructor: TextController });

});
